Package: CAST
Type: Package
Title: 'caret' Applications for Spatial-Temporal Models
Version: 1.1.0
Authors@R: c(person("Hanna", "Meyer", email = "hanna.meyer@uni-muenster.de", role = c("cre", "aut")),
             person("Carles", "Milà", role = c("aut")),
             person("Marvin", "Ludwig", role = c("aut")),
             person("Jan", "Linnenbrink", role = c("aut")),
             person("Fabian", "Schumacher", role = c("aut")),
             person("Philipp", "Otto", role = c("ctb")),
             person("Chris", "Reudenbach", role = c("ctb")),
             person("Thomas", "Nauss", role = c("ctb")),
             person("Edzer", "Pebesma", role = c("ctb")),
             person("Jakub", "Nowosad", role = c("ctb")),
             person("Darius", "Görgen", role = c("ctb")))
Description: Supporting functionality to run 'caret' with spatial or spatial-temporal data. 
    'caret' is a frequently used package for model training and prediction using machine learning. 
    CAST includes functions to improve spatial or spatial-temporal modelling tasks using 'caret'. 
    It includes the newly suggested 'Nearest neighbor distance matching' cross-validation to estimate
    the performance of spatial prediction models and allows for spatial variable selection to selects
    suitable predictor variables in view to their contribution to the spatial model performance. 
    CAST further includes functionality to estimate the (spatial) area of applicability of prediction models.
    Methods are described in 
    Meyer et al. (2018) <doi:10.1016/j.envsoft.2017.12.001>; 
    Meyer et al. (2019) <doi:10.1016/j.ecolmodel.2019.108815>; 
    Meyer and Pebesma (2021) <doi:10.1111/2041-210X.13650>; 
    Milà et al. (2022) <doi:10.1111/2041-210X.13851>; 
    Meyer and Pebesma (2022) <doi:10.1038/s41467-022-29838-9>; 
    Linnenbrink et al. (2024) <doi:10.5194/gmd-17-5897-2024>; 
    Schumacher et al. (2025) <doi:10.5194/gmd-18-10185-2025>. 
    The package is described in detail in Meyer et al. (2026) <doi:10.1007/978-3-031-99665-8_11>.
License: GPL (>= 2)
URL: https://github.com/HannaMeyer/CAST,
        https://hannameyer.github.io/CAST/
Encoding: UTF-8
LazyData: false
Depends: R (>= 4.1.0)
BugReports: https://github.com/HannaMeyer/CAST/issues/
Imports: caret, data.table, FNN, ggplot2, graphics, grDevices, methods,
        parallel, sf, stats, twosamples, terra, utils, zoo
Suggests: clustMixType, doParallel, geodata, gower, gridExtra, knitr,
        lubridate, MASS, PCAmixdata, randomForest, rmarkdown,
        rnaturalearth, scales, scam, stars, testthat (>= 3.0.0), tmap
VignetteBuilder: knitr
Config/testthat/edition: 3
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-06-19 13:21:16 UTC; hanna
Author: Hanna Meyer [cre, aut],
  Carles Milà [aut],
  Marvin Ludwig [aut],
  Jan Linnenbrink [aut],
  Fabian Schumacher [aut],
  Philipp Otto [ctb],
  Chris Reudenbach [ctb],
  Thomas Nauss [ctb],
  Edzer Pebesma [ctb],
  Jakub Nowosad [ctb],
  Darius Görgen [ctb]
Maintainer: Hanna Meyer <hanna.meyer@uni-muenster.de>
Repository: CRAN
Date/Publication: 2026-06-19 16:20:02 UTC
Built: R 4.5.2; ; 2026-06-19 20:20:52 UTC; unix
