# Generated by roxygen2: do not edit by hand

S3method(aoa,Raster)
S3method(aoa,SpatRaster)
S3method(aoa,data.frame)
S3method(aoa,stars)
S3method(plot,aoa)
S3method(plot,errorModel)
S3method(plot,ffs)
S3method(plot,geodist)
S3method(plot,knndm)
S3method(plot,nndm)
S3method(plot,trainDI)
S3method(print,aoa)
S3method(print,ffs)
S3method(print,knndm)
S3method(print,nndm)
S3method(print,trainDI)
export(CreateSpacetimeFolds)
export(aoa)
export(bss)
export(clustered_sample)
export(errorProfiles)
export(ffs)
export(geodist)
export(global_validation)
export(knndm)
export(nndm)
export(normalize_DI)
export(show.aoa)
export(show.ffs)
export(show.knndm)
export(show.nndm)
export(show.trainDI)
export(trainDI)
import(caret)
import(ggplot2)
importFrom(grDevices,rainbow)
importFrom(graphics,axis)
importFrom(graphics,boxplot)
importFrom(graphics,plot)
importFrom(graphics,segments)
importFrom(parallel,clusterExport)
importFrom(parallel,detectCores)
importFrom(parallel,makeForkCluster)
importFrom(parallel,parLapply)
importFrom(parallel,stopCluster)
importFrom(stats,complete.cases)
importFrom(stats,dist)
importFrom(stats,lm)
importFrom(stats,median)
importFrom(stats,na.exclude)
importFrom(stats,na.omit)
importFrom(stats,predict)
importFrom(stats,quantile)
importFrom(stats,sd)
importFrom(utils,combn)
importFrom(utils,setTxtProgressBar)
importFrom(utils,txtProgressBar)
